@@ -1,36 +1,74 @@
 package Sys::Syslog;
 use strict;
+use warnings::register;
 use Carp;
 use POSIX qw(strftime setlocale LC_TIME);
 use Socket ':all';
 require 5.006;
 require Exporter;
 
-our $VERSION = '0.14';
-our @ISA = qw(Exporter);
+{   no strict 'vars';
+    $VERSION = '0.15';
+    @ISA = qw(Exporter);
 
-our %EXPORT_TAGS = (
-    standard => [qw(openlog syslog closelog setlogmask)],
-    extended => [qw(setlogsock)],
-    macros => [qw(
-        LOG_ALERT LOG_AUTH LOG_AUTHPRIV LOG_CONS LOG_CRIT LOG_CRON
-        LOG_DAEMON LOG_DEBUG LOG_EMERG LOG_ERR LOG_FACMASK LOG_FTP
-        LOG_INFO LOG_KERN LOG_LFMT LOG_LOCAL0 LOG_LOCAL1 LOG_LOCAL2
-        LOG_LOCAL3 LOG_LOCAL4 LOG_LOCAL5 LOG_LOCAL6 LOG_LOCAL7 LOG_LPR
-        LOG_MAIL LOG_NDELAY LOG_NEWS LOG_NFACILITIES LOG_NOTICE
-        LOG_NOWAIT LOG_ODELAY LOG_PERROR LOG_PID LOG_PRIMASK LOG_SYSLOG
-        LOG_USER LOG_UUCP LOG_WARNING
-        LOG_MASK LOG_UPTO
-    )],
-);
+    %EXPORT_TAGS = (
+        standard => [qw(openlog syslog closelog setlogmask)],
+        extended => [qw(setlogsock)],
+        macros => [qw(
+            LOG_ALERT LOG_AUTH LOG_AUTHPRIV LOG_CONS LOG_CRIT LOG_CRON
+            LOG_DAEMON LOG_DEBUG LOG_EMERG LOG_ERR LOG_FACMASK LOG_FTP
+            LOG_INFO LOG_KERN LOG_LFMT LOG_LOCAL0 LOG_LOCAL1 LOG_LOCAL2
+            LOG_LOCAL3 LOG_LOCAL4 LOG_LOCAL5 LOG_LOCAL6 LOG_LOCAL7 LOG_LPR
+            LOG_MAIL LOG_NDELAY LOG_NEWS LOG_NFACILITIES LOG_NOTICE
+            LOG_NOWAIT LOG_ODELAY LOG_PERROR LOG_PID LOG_PRIMASK LOG_SYSLOG
+            LOG_USER LOG_UUCP LOG_WARNING
+            LOG_MASK LOG_UPTO
+        )],
+    );
 
-our @EXPORT = (
-    @{$EXPORT_TAGS{standard}}, 
-);
+    @EXPORT = (
+        @{$EXPORT_TAGS{standard}}, 
+    );
 
-our @EXPORT_OK = (
-    @{$EXPORT_TAGS{extended}}, 
-    @{$EXPORT_TAGS{macros}}, 
+    @EXPORT_OK = (
+        @{$EXPORT_TAGS{extended}}, 
+        @{$EXPORT_TAGS{macros}}, 
+    );
+
+    eval {
+        require XSLoader;
+        XSLoader::load('Sys::Syslog', $VERSION);
+        1
+    } or do {
+        require DynaLoader;
+        push @ISA, 'DynaLoader';
+        bootstrap Sys::Syslog $VERSION;
+    };
+}
+
+
+# 
+# Public variables
+# 
+our $host;                      # host to send syslog messages to
+
+# 
+# Global variables
+# 
+my $connected = 0;              # flag to indicate if we're connected or not
+my $syslog_send;                # coderef of the function used to send messages
+my $syslog_path = undef;        # syslog path for "stream" and "unix" mechanisms
+my $transmit_ok = 0;            # flag to indicate if the last message was transmited
+my $current_proto = undef;      # current mechanism used to transmit messages
+my $ident = '';                 # identifiant prepended to each message
+my $facility = '';              # current facility
+my $maskpri = LOG_UPTO(&LOG_DEBUG);     # current log mask
+
+my %options = (
+    ndelay  => 0, 
+    nofatal => 0, 
+    nowait  => 0, 
+    pid     => 0, 
 );
 
 # it would be nice to try stream/unix first, since that will be
@@ -40,13 +78,540 @@ if ($^O =~ /^(freebsd|linux)$/) {
     @connectMethods = grep { $_ ne 'udp' } @connectMethods;
 }
 my @defaultMethods = @connectMethods;
-my $syslog_path = undef;
-my $transmit_ok = 0;
-my $current_proto = undef;
-my $failed = undef;
-my $fail_time = undef;
-our ($connected, @fallbackMethods, $syslog_send, $host);
+my @fallbackMethods = ();
 
+# coderef for a nicer handling of errors
+my $err_sub = $options{nofatal} ? \&warnings::warnif : \&croak;
+
+
+sub AUTOLOAD {
+    # This AUTOLOAD is used to 'autoload' constants from the constant()
+    # XS function.
+    no strict 'vars';
+    my $constname;
+    ($constname = $AUTOLOAD) =~ s/.*:://;
+    croak "Sys::Syslog::constant() not defined" if $constname eq 'constant';
+    my ($error, $val) = constant($constname);
+	croak $error if $error;
+    no strict 'refs';
+    *$AUTOLOAD = sub { $val };
+    goto &$AUTOLOAD;
+}
+
+
+sub openlog {
+    ($ident, my $logopt, $facility) = @_;
+
+    for my $opt (split /\b/, $logopt) {
+        $options{$opt} = 1 if exists $options{$opt}
+    }
+
+    $err_sub = $options{nofatal} ? \&warnings::warnif : \&croak;
+    return 1 unless $options{ndelay};
+    connect_log();
+} 
+
+sub closelog {
+    $facility = $ident = '';
+    disconnect_log();
+} 
+
+sub setlogmask {
+    my $oldmask = $maskpri;
+    $maskpri = shift unless $_[0] == 0;
+    $oldmask;
+}
+ 
+sub setlogsock {
+    my $setsock = shift;
+    $syslog_path = shift;
+    disconnect_log() if $connected;
+    $transmit_ok = 0;
+    @fallbackMethods = ();
+    @connectMethods = @defaultMethods;
+
+    if (ref $setsock eq 'ARRAY') {
+	@connectMethods = @$setsock;
+
+    } elsif (lc $setsock eq 'stream') {
+	unless (defined $syslog_path) {
+	    my @try = qw(/dev/log /dev/conslog);
+	    if (length &_PATH_LOG) { # Undefined _PATH_LOG is "".
+		unshift @try, &_PATH_LOG;
+            }
+	    for my $try (@try) {
+		if (-w $try) {
+		    $syslog_path = $try;
+		    last;
+		}
+	    }
+            warnings::warnif "stream passed to setlogsock, but could not find any device"
+		unless defined $syslog_path
+        }
+	unless (-w $syslog_path) {
+        warnings::warnif "stream passed to setlogsock, but $syslog_path is not writable";
+	    return undef;
+	} else {
+	    @connectMethods = ( 'stream' );
+	}
+
+    } elsif (lc $setsock eq 'unix') {
+        if (length _PATH_LOG() && !defined $syslog_path) {
+	    $syslog_path = _PATH_LOG();
+	    @connectMethods = ( 'unix' );
+        } else {
+            warnings::warnif 'unix passed to setlogsock, but path not available';
+	    return undef;
+        }
+
+    } elsif (lc $setsock eq 'native') {
+        @connectMethods = ( 'native' );
+
+    } elsif (lc $setsock eq 'tcp') {
+	if (getservbyname('syslog', 'tcp') || getservbyname('syslogng', 'tcp')) {
+	    @connectMethods = ( 'tcp' );
+	} else {
+            warnings::warnif "tcp passed to setlogsock, but tcp service unavailable";
+	    return undef;
+	}
+
+    } elsif (lc $setsock eq 'udp') {
+	if (getservbyname('syslog', 'udp')) {
+	    @connectMethods = ( 'udp' );
+	} else {
+            warnings::warnif "udp passed to setlogsock, but udp service unavailable";
+	    return undef;
+	}
+
+    } elsif (lc $setsock eq 'inet') {
+	@connectMethods = ( 'tcp', 'udp' );
+
+    } elsif (lc $setsock eq 'console') {
+	@connectMethods = ( 'console' );
+
+    } else {
+        croak "Invalid argument passed to setlogsock; must be 'stream', 'unix', 'native', 'tcp', 'udp' or 'inet'"
+    }
+
+    return 1;
+}
+
+sub syslog {
+    my $priority = shift;
+    my $mask = shift;
+    my ($message, $buf);
+    my (@words, $num, $numpri, $numfac, $sum);
+    my $failed = undef;
+    my $fail_time = undef;
+
+    my $facility = $facility;	# may need to change temporarily.
+
+    croak "syslog: expecting argument \$priority" unless defined $priority;
+    croak "syslog: expecting argument \$format"   unless defined $mask;
+
+    @words = split(/\W+/, $priority, 2);# Allow "level" or "level|facility".
+    undef $numpri;
+    undef $numfac;
+
+    foreach (@words) {
+	$num = xlate($_);		# Translate word to number.
+	if ($num < 0) {
+	    croak "syslog: invalid level/facility: $_"
+	}
+	elsif ($num <= &LOG_PRIMASK) {
+	    croak "syslog: too many levels given: $_" if defined $numpri;
+	    $numpri = $num;
+	    return 0 unless LOG_MASK($numpri) & $maskpri;
+	}
+	else {
+	    croak "syslog: too many facilities given: $_" if defined $numfac;
+	    $facility = $_;
+	    $numfac = $num;
+	}
+    }
+
+    croak "syslog: level must be given" unless defined $numpri;
+
+    if (not defined $numfac) {  # Facility not specified in this call.
+	$facility = 'user' unless $facility;
+	$numfac = xlate($facility);
+    }
+
+    connect_log() unless $connected;
+
+    if ($mask =~ /%m/) {
+	my $err = $!;
+	# escape percent signs if sprintf will be called
+	$err =~ s/%/%%/g if @_;
+	# replace %m with $err, if preceded by an even number of percent signs
+	$mask =~ s/(?<!%)((?:%%)*)%m/$1$err/g;
+    }
+
+    $mask .= "\n" unless $mask =~ /\n$/;
+    $message = @_ ? sprintf($mask, @_) : $mask;
+
+    if($current_proto eq 'native') {
+        $buf = $message;
+
+    } else {
+        my $whoami = $ident;
+
+        if (!$whoami && $mask =~ /^(\S.*?):\s?(.*)/) {
+            $whoami = $1;
+            $mask = $2;
+        }
+
+        unless ($whoami) {
+            $whoami = getlogin() || getpwuid($<) || 'syslog';
+        }
+
+        $whoami .= "[$$]" if $options{pid};
+
+        $sum = $numpri + $numfac;
+        my $oldlocale = setlocale(LC_TIME);
+        setlocale(LC_TIME, 'C');
+        my $timestamp = strftime "%b %e %T", localtime;
+        setlocale(LC_TIME, $oldlocale);
+        $buf = "<$sum>$timestamp $whoami: $message\0";
+    }
+
+    # it's possible that we'll get an error from sending
+    # (e.g. if method is UDP and there is no UDP listener,
+    # then we'll get ECONNREFUSED on the send). So what we
+    # want to do at this point is to fallback onto a different
+    # connection method.
+    while (scalar @fallbackMethods || $syslog_send) {
+	if ($failed && (time - $fail_time) > 60) {
+	    # it's been a while... maybe things have been fixed
+	    @fallbackMethods = ();
+	    disconnect_log();
+	    $transmit_ok = 0; # make it look like a fresh attempt
+	    connect_log();
+        }
+
+	if ($connected && !connection_ok()) {
+	    # Something was OK, but has now broken. Remember coz we'll
+	    # want to go back to what used to be OK.
+	    $failed = $current_proto unless $failed;
+	    $fail_time = time;
+	    disconnect_log();
+	}
+
+	connect_log() unless $connected;
+	$failed = undef if ($current_proto && $failed && $current_proto eq $failed);
+
+	if ($syslog_send) {
+            if ($syslog_send->($buf, $numpri)) {
+		$transmit_ok++;
+		return 1;
+	    }
+	    # typically doesn't happen, since errors are rare from write().
+	    disconnect_log();
+	}
+    }
+    # could not send, could not fallback onto a working
+    # connection method. Lose.
+    return 0;
+}
+
+sub _syslog_send_console {
+    my ($buf) = @_;
+    chop($buf); # delete the NUL from the end
+    # The console print is a method which could block
+    # so we do it in a child process and always return success
+    # to the caller.
+    if (my $pid = fork) {
+
+	if ($options{nowait}) {
+	    return 1;
+	} else {
+	    if (waitpid($pid, 0) >= 0) {
+	    	return ($? >> 8);
+	    } else {
+		# it's possible that the caller has other
+		# plans for SIGCHLD, so let's not interfere
+		return 1;
+	    }
+	}
+    } else {
+        if (open(CONS, ">/dev/console")) {
+	    my $ret = print CONS $buf . "\r";  # XXX: should this be \x0A ?
+	    exit $ret if defined $pid;
+	    close CONS;
+	}
+	exit if defined $pid;
+    }
+}
+
+sub _syslog_send_stream {
+    my ($buf) = @_;
+    # XXX: this only works if the OS stream implementation makes a write 
+    # look like a putmsg() with simple header. For instance it works on 
+    # Solaris 8 but not Solaris 7.
+    # To be correct, it should use a STREAMS API, but perl doesn't have one.
+    return syswrite(SYSLOG, $buf, length($buf));
+}
+
+sub _syslog_send_socket {
+    my ($buf) = @_;
+    return syswrite(SYSLOG, $buf, length($buf));
+    #return send(SYSLOG, $buf, 0);
+}
+
+sub _syslog_send_native {
+    my ($buf, $numpri) = @_;
+    eval { syslog_xs($numpri, $buf) };
+    return $@ ? 0 : 1;
+}
+
+
+# xlate()
+# -----
+# private function to translate names to numeric values
+# 
+sub xlate {
+    my($name) = @_;
+    return $name+0 if $name =~ /^\s*\d+\s*$/;
+    $name = uc $name;
+    $name = "LOG_$name" unless $name =~ /^LOG_/;
+    $name = "Sys::Syslog::$name";
+    # Can't have just eval { &$name } || -1 because some LOG_XXX may be zero.
+    my $value = eval { no strict 'refs'; &$name };
+    defined $value ? $value : -1;
+}
+
+
+# connect_log()
+# -----------
+# This function acts as a kind of front-end: it tries to connect to 
+# a syslog service using the selected methods, trying each one in the 
+# selected order. 
+# 
+sub connect_log {
+    @fallbackMethods = @connectMethods unless scalar @fallbackMethods;
+    if ($transmit_ok && $current_proto) {
+        # Retry what we were on, because it has worked in the past.
+	unshift(@fallbackMethods, $current_proto);
+    }
+    $connected = 0;
+    my @errs = ();
+    my $proto = undef;
+    while ($proto = shift @fallbackMethods) {
+	no strict 'refs';
+	my $fn = "connect_$proto";
+	$connected = &$fn(\@errs) if defined &$fn;
+	last if $connected;
+    }
+
+    $transmit_ok = 0;
+    if ($connected) {
+	$current_proto = $proto;
+        my($old) = select(SYSLOG); $| = 1; select($old);
+    } else {
+	@fallbackMethods = ();
+        $err_sub->(join "\n\t- ", "no connection to syslog available", @errs);
+        return undef;
+    }
+}
+
+sub connect_tcp {
+    my ($errs) = @_;
+    my $tcp = getprotobyname('tcp');
+    if (!defined $tcp) {
+	push @$errs, "getprotobyname failed for tcp";
+	return 0;
+    }
+    my $syslog = getservbyname('syslog','tcp');
+    $syslog = getservbyname('syslogng','tcp') unless defined $syslog;
+    if (!defined $syslog) {
+	push @$errs, "getservbyname failed for syslog/tcp and syslogng/tcp";
+	return 0;
+    }
+
+    my $this = sockaddr_in($syslog, INADDR_ANY);
+    my $that;
+    if (defined $host) {
+	$that = inet_aton($host);
+	if (!$that) {
+	    push @$errs, "can't lookup $host";
+	    return 0;
+	}
+    } else {
+	$that = INADDR_LOOPBACK;
+    }
+    $that = sockaddr_in($syslog, $that);
+
+    if (!socket(SYSLOG, AF_INET, SOCK_STREAM, $tcp)) {
+	push @$errs, "tcp socket: $!";
+	return 0;
+    }
+    setsockopt(SYSLOG, SOL_SOCKET, SO_KEEPALIVE, 1);
+    setsockopt(SYSLOG, &IPPROTO_TCP, &TCP_NODELAY, 1);
+    if (!connect(SYSLOG, $that)) {
+	push @$errs, "tcp connect: $!";
+	return 0;
+    }
+    $syslog_send = \&_syslog_send_socket;
+    return 1;
+}
+
+sub connect_udp {
+    my ($errs) = @_;
+    my $udp = getprotobyname('udp');
+    if (!defined $udp) {
+	push @$errs, "getprotobyname failed for udp";
+	return 0;
+    }
+    my $syslog = getservbyname('syslog','udp');
+    if (!defined $syslog) {
+	push @$errs, "getservbyname failed for syslog/udp";
+	return 0;
+    }
+    my $this = sockaddr_in($syslog, INADDR_ANY);
+    my $that;
+    if (defined $host) {
+	$that = inet_aton($host);
+	if (!$that) {
+	    push @$errs, "can't lookup $host";
+	    return 0;
+	}
+    } else {
+	$that = INADDR_LOOPBACK;
+    }
+    $that = sockaddr_in($syslog, $that);
+
+    if (!socket(SYSLOG, AF_INET, SOCK_DGRAM, $udp)) {
+	push @$errs, "udp socket: $!";
+	return 0;
+    }
+    if (!connect(SYSLOG, $that)) {
+	push @$errs, "udp connect: $!";
+	return 0;
+    }
+    # We want to check that the UDP connect worked. However the only
+    # way to do that is to send a message and see if an ICMP is returned
+    _syslog_send_socket("");
+    if (!connection_ok()) {
+	push @$errs, "udp connect: nobody listening";
+	return 0;
+    }
+    $syslog_send = \&_syslog_send_socket;
+    return 1;
+}
+
+sub connect_stream {
+    my ($errs) = @_;
+    # might want syslog_path to be variable based on syslog.h (if only
+    # it were in there!)
+    $syslog_path = '/dev/conslog'; 
+    if (!-w $syslog_path) {
+	push @$errs, "stream $syslog_path is not writable";
+	return 0;
+    }
+    if (!open(SYSLOG, ">" . $syslog_path)) {
+	push @$errs, "stream can't open $syslog_path: $!";
+	return 0;
+    }
+    $syslog_send = \&_syslog_send_stream;
+    return 1;
+}
+
+sub connect_unix {
+    my ($errs) = @_;
+    if (length _PATH_LOG()) {
+	$syslog_path = _PATH_LOG();
+    } else {
+        push @$errs, "_PATH_LOG not available in syslog.h";
+	return 0;
+    }
+    if (! -S $syslog_path) {
+        push @$errs, "$syslog_path is not a socket";
+	return 0;
+    }
+    my $that = sockaddr_un($syslog_path);
+    if (!$that) {
+	push @$errs, "can't locate $syslog_path";
+	return 0;
+    }
+    if (!socket(SYSLOG,AF_UNIX,SOCK_STREAM,0)) {
+        push @$errs, "unix stream socket: $!";
+	return 0;
+    }
+    if (!connect(SYSLOG,$that)) {
+        if (!socket(SYSLOG,AF_UNIX,SOCK_DGRAM,0)) {
+	    push @$errs, "unix dgram socket: $!";
+	    return 0;
+	}
+        if (!connect(SYSLOG,$that)) {
+	    push @$errs, "unix dgram connect: $!";
+	    return 0;
+	}
+    }
+    $syslog_send = \&_syslog_send_socket;
+    return 1;
+}
+
+sub connect_native {
+    my ($errs) = @_;
+    my $logopt = 0;
+
+    # reconstruct the numeric equivalent of the options
+    for my $opt (keys %options) {
+        $logopt += xlate($opt) if $options{$opt}
+    }
+
+    eval { openlog_xs($ident, $logopt, xlate($facility)) };
+    if ($@) {
+        push @$errs, $@;
+        return 0;
+    }
+
+    $syslog_send = \&_syslog_send_native;
+
+    return 1;
+}
+
+sub connect_console {
+    my ($errs) = @_;
+    if (!-w '/dev/console') {
+	push @$errs, "console is not writable";
+	return 0;
+    }
+    $syslog_send = \&_syslog_send_console;
+    return 1;
+}
+
+# to test if the connection is still good, we need to check if any
+# errors are present on the connection. The errors will not be raised
+# by a write. Instead, sockets are made readable and the next read
+# would cause the error to be returned. Unfortunately the syslog 
+# 'protocol' never provides anything for us to read. But with 
+# judicious use of select(), we can see if it would be readable...
+sub connection_ok {
+    return 1 if defined $current_proto and (
+        $current_proto eq 'native' or $current_proto eq 'console'
+    );
+    my $rin = '';
+    vec($rin, fileno(SYSLOG), 1) = 1;
+    my $ret = select $rin, undef, $rin, 0;
+    return ($ret ? 0 : 1);
+}
+
+sub disconnect_log {
+    $connected = 0;
+    $syslog_send = undef;
+
+    if($current_proto eq 'native') {
+        eval { close_xs() };
+        return 1;
+    }
+
+    return close SYSLOG;
+}
+
+1;
+
+__END__
 
 =head1 NAME
 
@@ -54,7 +619,7 @@ Sys::Syslog - Perl interface to the UNIX syslog(3) calls
 
 =head1 VERSION
 
-Version 0.14
+Version 0.15
 
 =head1 SYNOPSIS
 
@@ -113,18 +678,18 @@ By default, C<Sys::Syslog> exports the symbols from the C<:standard> tag.
 
 Opens the syslog.
 C<$ident> is prepended to every message.  C<$logopt> contains zero or
-more of the words C<pid>, C<ndelay>, C<nowait>.  The C<cons> option is
-ignored, since the failover mechanism will drop down to the console
-automatically if all other media fail.  C<$facility> specifies the
-part of the system to report about, for example C<LOG_USER> or C<LOG_LOCAL0>:
-see your C<syslog(3)> documentation for the facilities available in
-your system. Facility can be given as a string or a numeric macro. 
+more of the options detailed below.  C<$facility> specifies the part 
+of the system to report about, for example C<LOG_USER> or C<LOG_LOCAL0>:
+see L<"Facilities"> for a list of well-known facilities, and your 
+C<syslog(3)> documentation for the facilities available in your system. 
+Check L<"SEE ALSO"> for useful links. Facility can be given as a string 
+or a numeric macro. 
 
 This function will croak if it can't connect to the syslog daemon.
 
 Note that C<openlog()> now takes three arguments, just like C<openlog(3)>.
 
-B<You should use openlog() before calling syslog().>
+B<You should use C<openlog()> before calling C<syslog()>.>
 
 B<Options>
 
@@ -132,11 +697,22 @@ B<Options>
 
 =item *
 
+C<cons> - This option is ignored, since the failover mechanism will drop 
+down to the console automatically if all other media fail.
+
+=item *
+
 C<ndelay> - Open the connection immediately (normally, the connection is
 opened when the first message is logged).
 
 =item *
 
+C<nofatal> - When set to true, C<openlog()> and C<syslog()> will only 
+emit warnings instead of dying if the connection to the syslog can't 
+be established. 
+
+=item *
+
 C<nowait> - Don't wait for child processes that may have been created 
 while logging the message.  (The GNU C library does not create a child
 process, so this option has no effect on Linux.)
@@ -495,6 +1071,9 @@ Code for C<constant()>s regenerated by Nicholas Clark E<lt>F<nick@ccl4.org>E<gt>
 Failover to different communication modes by Nick Williams
 E<lt>F<Nick.Williams@morganstanley.com>E<gt>.
 
+XS code for using native C functions borrowed from C<L<Unix::Syslog>>, 
+written by Marcus Harnisch E<lt>F<marcus.harnisch@gmx.net>E<gt>.
+
 Extracted from core distribution for publishing on the CPAN by 
 SE<eacute>bastien Aperghis-Tramoni E<lt>sebastien@aperghis.netE<gt>.
 
@@ -551,485 +1130,3 @@ This program is free software; you can redistribute it and/or modify it
 under the same terms as Perl itself.
 
 =cut
-
-sub AUTOLOAD {
-    # This AUTOLOAD is used to 'autoload' constants from the constant()
-    # XS function.
-    my $constname;
-    our $AUTOLOAD;
-    ($constname = $AUTOLOAD) =~ s/.*:://;
-    croak "Sys::Syslog::constant() not defined" if $constname eq 'constant';
-    my ($error, $val) = constant($constname);
-	croak $error if $error;
-    no strict 'refs';
-    *$AUTOLOAD = sub { $val };
-    goto &$AUTOLOAD;
-}
-
-eval {
-    require XSLoader;
-    XSLoader::load('Sys::Syslog', $VERSION);
-    1
-} or do {
-    require DynaLoader;
-    push @ISA, 'DynaLoader';
-    bootstrap Sys::Syslog $VERSION;
-};
-
-our $maskpri = LOG_UPTO(&LOG_DEBUG);
-
-sub openlog {
-    our ($ident, $logopt, $facility) = @_;  # package vars
-    our $lo_pid = $logopt =~ /\bpid\b/;
-    our $lo_ndelay = $logopt =~ /\bndelay\b/;
-    our $lo_nowait = $logopt =~ /\bnowait\b/;
-    return 1 unless $lo_ndelay;
-    connect_log();
-} 
-
-sub closelog {
-    our $facility = our $ident = '';
-    disconnect_log();
-} 
-
-sub setlogmask {
-    my $oldmask = $maskpri;
-    $maskpri = shift unless $_[0] == 0;
-    $oldmask;
-}
- 
-sub setlogsock {
-    my $setsock = shift;
-    $syslog_path = shift;
-    disconnect_log() if $connected;
-    $transmit_ok = 0;
-    @fallbackMethods = ();
-    @connectMethods = @defaultMethods;
-
-    if (ref $setsock eq 'ARRAY') {
-	@connectMethods = @$setsock;
-
-    } elsif (lc($setsock) eq 'stream') {
-	unless (defined $syslog_path) {
-	    my @try = qw(/dev/log /dev/conslog);
-	    if (length &_PATH_LOG) { # Undefined _PATH_LOG is "".
-		unshift @try, &_PATH_LOG;
-            }
-	    for my $try (@try) {
-		if (-w $try) {
-		    $syslog_path = $try;
-		    last;
-		}
-	    }
-	    carp "stream passed to setlogsock, but could not find any device"
-		unless defined $syslog_path
-        }
-	unless (-w $syslog_path) {
-	    carp "stream passed to setlogsock, but $syslog_path is not writable";
-	    return undef;
-	} else {
-	    @connectMethods = ( 'stream' );
-	}
-
-    } elsif (lc($setsock) eq 'unix') {
-        if (length _PATH_LOG() && !defined $syslog_path) {
-	    $syslog_path = _PATH_LOG();
-	    @connectMethods = ( 'unix' );
-        } else {
-	    carp 'unix passed to setlogsock, but path not available';
-	    return undef;
-        }
-
-    } elsif (lc($setsock) eq 'tcp') {
-	if (getservbyname('syslog', 'tcp') || getservbyname('syslogng', 'tcp')) {
-	    @connectMethods = ( 'tcp' );
-	} else {
-	    carp "tcp passed to setlogsock, but tcp service unavailable";
-	    return undef;
-	}
-
-    } elsif (lc($setsock) eq 'udp') {
-	if (getservbyname('syslog', 'udp')) {
-	    @connectMethods = ( 'udp' );
-	} else {
-	    carp "udp passed to setlogsock, but udp service unavailable";
-	    return undef;
-	}
-
-    } elsif (lc($setsock) eq 'inet') {
-	@connectMethods = ( 'tcp', 'udp' );
-
-    } elsif (lc($setsock) eq 'console') {
-	@connectMethods = ( 'console' );
-
-    } else {
-        croak "Invalid argument passed to setlogsock; must be 'stream', 'unix', 'tcp', 'udp' or 'inet'"
-    }
-
-    return 1;
-}
-
-sub syslog {
-    my $priority = shift;
-    my $mask = shift;
-    my ($message, $whoami);
-    my (@words, $num, $numpri, $numfac, $sum);
-    our $facility;
-    local($facility) = $facility;	# may need to change temporarily.
-
-    croak "syslog: expecting argument \$priority" unless defined $priority;
-    croak "syslog: expecting argument \$format"   unless defined $mask;
-
-    @words = split(/\W+/, $priority, 2);# Allow "level" or "level|facility".
-    undef $numpri;
-    undef $numfac;
-
-    foreach (@words) {
-	$num = xlate($_);		# Translate word to number.
-	if ($num < 0) {
-	    croak "syslog: invalid level/facility: $_"
-	}
-	elsif ($num <= &LOG_PRIMASK) {
-	    croak "syslog: too many levels given: $_" if defined($numpri);
-	    $numpri = $num;
-	    return 0 unless LOG_MASK($numpri) & $maskpri;
-	}
-	else {
-	    croak "syslog: too many facilities given: $_" if defined($numfac);
-	    $facility = $_;
-	    $numfac = $num;
-	}
-    }
-
-    croak "syslog: level must be given" unless defined($numpri);
-
-    if (!defined($numfac)) {	# Facility not specified in this call.
-	$facility = 'user' unless $facility;
-	$numfac = xlate($facility);
-    }
-
-    connect_log() unless $connected;
-
-    $whoami = our $ident;
-
-    if (!$whoami && $mask =~ /^(\S.*?):\s?(.*)/) {
-	$whoami = $1;
-	$mask = $2;
-    } 
-
-    unless ($whoami) {
-	$whoami = getlogin() || getpwuid($<) || 'syslog';
-    }
-
-    $whoami .= "[$$]" if our $lo_pid;
-
-    if ($mask =~ /%m/) {
-	my $err = $!;
-	# escape percent signs if sprintf will be called
-	$err =~ s/%/%%/g if @_;
-	# replace %m with $err, if preceded by an even number of percent signs
-	$mask =~ s/(?<!%)((?:%%)*)%m/$1$err/g;
-    }
-
-    $mask .= "\n" unless $mask =~ /\n$/;
-    $message = @_ ? sprintf($mask, @_) : $mask;
-
-    $sum = $numpri + $numfac;
-    my $oldlocale = setlocale(LC_TIME);
-    setlocale(LC_TIME, 'C');
-    my $timestamp = strftime "%b %e %T", localtime;
-    setlocale(LC_TIME, $oldlocale);
-    my $buf = "<$sum>$timestamp $whoami: $message\0";
-
-    # it's possible that we'll get an error from sending
-    # (e.g. if method is UDP and there is no UDP listener,
-    # then we'll get ECONNREFUSED on the send). So what we
-    # want to do at this point is to fallback onto a different
-    # connection method.
-    while (scalar @fallbackMethods || $syslog_send) {
-	if ($failed && (time - $fail_time) > 60) {
-	    # it's been a while... maybe things have been fixed
-	    @fallbackMethods = ();
-	    disconnect_log();
-	    $transmit_ok = 0; # make it look like a fresh attempt
-	    connect_log();
-        }
-
-	if ($connected && !connection_ok()) {
-	    # Something was OK, but has now broken. Remember coz we'll
-	    # want to go back to what used to be OK.
-	    $failed = $current_proto unless $failed;
-	    $fail_time = time;
-	    disconnect_log();
-	}
-
-	connect_log() unless $connected;
-	$failed = undef if ($current_proto && $failed && $current_proto eq $failed);
-
-	if ($syslog_send) {
-	    if ($syslog_send->($buf)) {
-		$transmit_ok++;
-		return 1;
-	    }
-	    # typically doesn't happen, since errors are rare from write().
-	    disconnect_log();
-	}
-    }
-    # could not send, could not fallback onto a working
-    # connection method. Lose.
-    return 0;
-}
-
-sub _syslog_send_console {
-    my ($buf) = @_;
-    chop($buf); # delete the NUL from the end
-    # The console print is a method which could block
-    # so we do it in a child process and always return success
-    # to the caller.
-    if (my $pid = fork) {
-	our $lo_nowait;
-	if ($lo_nowait) {
-	    return 1;
-	} else {
-	    if (waitpid($pid, 0) >= 0) {
-	    	return ($? >> 8);
-	    } else {
-		# it's possible that the caller has other
-		# plans for SIGCHLD, so let's not interfere
-		return 1;
-	    }
-	}
-    } else {
-        if (open(CONS, ">/dev/console")) {
-	    my $ret = print CONS $buf . "\r";
-	    exit ($ret) if defined $pid;
-	    close CONS;
-	}
-	exit if defined $pid;
-    }
-}
-
-sub _syslog_send_stream {
-    my ($buf) = @_;
-    # XXX: this only works if the OS stream implementation makes a write 
-    # look like a putmsg() with simple header. For instance it works on 
-    # Solaris 8 but not Solaris 7.
-    # To be correct, it should use a STREAMS API, but perl doesn't have one.
-    return syswrite(SYSLOG, $buf, length($buf));
-}
-
-sub _syslog_send_socket {
-    my ($buf) = @_;
-    return syswrite(SYSLOG, $buf, length($buf));
-    #return send(SYSLOG, $buf, 0);
-}
-
-# xlate()
-# -----
-# private function to translate names to numeric values
-# 
-sub xlate {
-    my($name) = @_;
-    return $name+0 if $name =~ /^\s*\d+\s*$/;
-    $name = uc $name;
-    $name = "LOG_$name" unless $name =~ /^LOG_/;
-    $name = "Sys::Syslog::$name";
-    # Can't have just eval { &$name } || -1 because some LOG_XXX may be zero.
-    my $value = eval { no strict 'refs'; &$name };
-    defined $value ? $value : -1;
-}
-
-sub connect_log {
-    @fallbackMethods = @connectMethods unless (scalar @fallbackMethods);
-    if ($transmit_ok && $current_proto) {
-	# Retry what we were on, because it's worked in the past.
-	unshift(@fallbackMethods, $current_proto);
-    }
-    $connected = 0;
-    my @errs = ();
-    my $proto = undef;
-    while ($proto = shift(@fallbackMethods)) {
-	no strict 'refs';
-	my $fn = "connect_$proto";
-	$connected = &$fn(\@errs) if defined &$fn;
-	last if ($connected);
-    }
-
-    $transmit_ok = 0;
-    if ($connected) {
-	$current_proto = $proto;
-        my($old) = select(SYSLOG); $| = 1; select($old);
-    } else {
-	@fallbackMethods = ();
-	croak join "\n\t- ", "no connection to syslog available", @errs
-    }
-}
-
-sub connect_tcp {
-    my ($errs) = @_;
-    my $tcp = getprotobyname('tcp');
-    if (!defined $tcp) {
-	push(@{$errs}, "getprotobyname failed for tcp");
-	return 0;
-    }
-    my $syslog = getservbyname('syslog','tcp');
-    $syslog = getservbyname('syslogng','tcp') unless (defined $syslog);
-    if (!defined $syslog) {
-	push(@{$errs}, "getservbyname failed for syslog/tcp and syslogng/tcp");
-	return 0;
-    }
-
-    my $this = sockaddr_in($syslog, INADDR_ANY);
-    my $that;
-    if (defined $host) {
-	$that = inet_aton($host);
-	if (!$that) {
-	    push(@{$errs}, "can't lookup $host");
-	    return 0;
-	}
-    } else {
-	$that = INADDR_LOOPBACK;
-    }
-    $that = sockaddr_in($syslog, $that);
-
-    if (!socket(SYSLOG,AF_INET,SOCK_STREAM,$tcp)) {
-	push(@{$errs}, "tcp socket: $!");
-	return 0;
-    }
-    setsockopt(SYSLOG, SOL_SOCKET, SO_KEEPALIVE, 1);
-    setsockopt(SYSLOG, IPPROTO_TCP, TCP_NODELAY, 1);
-    if (!connect(SYSLOG,$that)) {
-	push(@{$errs}, "tcp connect: $!");
-	return 0;
-    }
-    $syslog_send = \&_syslog_send_socket;
-    return 1;
-}
-
-sub connect_udp {
-    my ($errs) = @_;
-    my $udp = getprotobyname('udp');
-    if (!defined $udp) {
-	push(@{$errs}, "getprotobyname failed for udp");
-	return 0;
-    }
-    my $syslog = getservbyname('syslog','udp');
-    if (!defined $syslog) {
-	push(@{$errs}, "getservbyname failed for syslog/udp");
-	return 0;
-    }
-    my $this = sockaddr_in($syslog, INADDR_ANY);
-    my $that;
-    if (defined $host) {
-	$that = inet_aton($host);
-	if (!$that) {
-	    push(@{$errs}, "can't lookup $host");
-	    return 0;
-	}
-    } else {
-	$that = INADDR_LOOPBACK;
-    }
-    $that = sockaddr_in($syslog, $that);
-
-    if (!socket(SYSLOG,AF_INET,SOCK_DGRAM,$udp)) {
-	push(@{$errs}, "udp socket: $!");
-	return 0;
-    }
-    if (!connect(SYSLOG,$that)) {
-	push(@{$errs}, "udp connect: $!");
-	return 0;
-    }
-    # We want to check that the UDP connect worked. However the only
-    # way to do that is to send a message and see if an ICMP is returned
-    _syslog_send_socket("");
-    if (!connection_ok()) {
-	push(@{$errs}, "udp connect: nobody listening");
-	return 0;
-    }
-    $syslog_send = \&_syslog_send_socket;
-    return 1;
-}
-
-sub connect_stream {
-    my ($errs) = @_;
-    # might want syslog_path to be variable based on syslog.h (if only
-    # it were in there!)
-    $syslog_path = '/dev/conslog'; 
-    if (!-w $syslog_path) {
-	push(@{$errs}, "stream $syslog_path is not writable");
-	return 0;
-    }
-    if (!open(SYSLOG, ">" . $syslog_path)) {
-	push(@{$errs}, "stream can't open $syslog_path: $!");
-	return 0;
-    }
-    $syslog_send = \&_syslog_send_stream;
-    return 1;
-}
-
-sub connect_unix {
-    my ($errs) = @_;
-    if (length _PATH_LOG()) {
-	$syslog_path = _PATH_LOG();
-    } else {
-        push(@{$errs}, "_PATH_LOG not available in syslog.h");
-	return 0;
-    }
-    if (! -S $syslog_path) {
-	push(@{$errs}, "$syslog_path is not a socket");
-	return 0;
-    }
-    my $that = sockaddr_un($syslog_path);
-    if (!$that) {
-	push(@{$errs}, "can't locate $syslog_path");
-	return 0;
-    }
-    if (!socket(SYSLOG,AF_UNIX,SOCK_STREAM,0)) {
-	push(@{$errs}, "unix stream socket: $!");
-	return 0;
-    }
-    if (!connect(SYSLOG,$that)) {
-        if (!socket(SYSLOG,AF_UNIX,SOCK_DGRAM,0)) {
-	    push(@{$errs}, "unix dgram socket: $!");
-	    return 0;
-	}
-        if (!connect(SYSLOG,$that)) {
-	    push(@{$errs}, "unix dgram connect: $!");
-	    return 0;
-	}
-    }
-    $syslog_send = \&_syslog_send_socket;
-    return 1;
-}
-
-sub connect_console {
-    my ($errs) = @_;
-    if (!-w '/dev/console') {
-	push(@{$errs}, "console is not writable");
-	return 0;
-    }
-    $syslog_send = \&_syslog_send_console;
-    return 1;
-}
-
-# to test if the connection is still good, we need to check if any
-# errors are present on the connection. The errors will not be raised
-# by a write. Instead, sockets are made readable and the next read
-# would cause the error to be returned. Unfortunately the syslog 
-# 'protocol' never provides anything for us to read. But with 
-# judicious use of select(), we can see if it would be readable...
-sub connection_ok {
-    return 1 if (defined $current_proto && $current_proto eq 'console');
-    my $rin = '';
-    vec($rin, fileno(SYSLOG), 1) = 1;
-    my $ret = select $rin, undef, $rin, 0;
-    return ($ret ? 0 : 1);
-}
-
-sub disconnect_log {
-    $connected = 0;
-    $syslog_send = undef;
-    return close SYSLOG;
-}
-
-1;
